/* Searches using  Horspool's algorithm adapted to 
search inside a text compressed with End-Tagged Dense Code.
Lightweight Natural Language Text Compression: Information Retrieval 2006

Programmed by Antonio Faria.

Author's contact: Antonio Faria, Databases Lab, University of
A Corua. Campus de Elvia s/n. Spain  fari@udc.es

Copyright (C) 2006  Nieves R. Brisaboa, Gonzalo Navarro, Antonio Faria and Jos R. Param
Author's contact: antonio.fari@gmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/*-----------------------------------------------------------------------
  RandomSearch.c:     Main file of the searcher.
  Loads the vocabulary of the compressor into a hash table. 
  Then it loads a pattern for each iteration of the search process).
  The pattern is searched inside the compressed file.  
  A file containing random patterns can be created by using: 
  		2.ExtractWords and 3.GenerateRandomPatterns programs.
  		
  	 The format of that file is   "%s\n"  (a pattern in each line)    		
 ------------------------------------------------------------------------*/
 
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <stdarg.h>
#include <stdlib.h>

#include <stdarg.h>  //para el vscanf

#include "tiempo.h"
#include "hash.h"
#include "searcher.c"

struct Pats {
        unsigned char   word[MAX_SIZE_OF_WORD];   // word from vocabulary
        unsigned long 	numBytesWord;
} ;
typedef struct Pats t_Pats;

t_Pats nextwordpattern;

FILE *ftiempos;   //file for storing search times.
FILE *fPatterns;  //file containing the patterns.


/*-----------------------------------------------------------------------
 Loads the vocabulary used by the decompressor into a hash table. The key of the hash table 
 is the word,  and the value --> the relative position in the ordered vocabulary.
 (or in this case, the codeword associated to it).
 ------------------------------------------------------------------------*/
void loadCompressorHashTable(char * fileName) {
	BufferRead V;	
	int v;
	unsigned char *text,*top, *nextbuf, *pbeg, *pend, *codeStart;
	char *voc;
	
	voc = (char *) malloc(strlen(fileName)+256);
	strcpy(voc,fileName);
	strcat(voc,".voc");	
	
	// Inicialization of the buffers used to read the files

	v = open (voc, O_RDONLY);
	if (v == -1) {
		fprintf(stderr,"Cannot read vocabulary file %s\nSearch aborted.\n",voc);
		exit(0);
	}	
 
	V = bufCreate(MAX_SIZE_OF_WORD+5);	  //a word being "cut"
	bufSetFile (V,v);
	
	/* loading the vocabulary into memory*/
	bufCurrent (V,(byte**)&text,(byte**)&top);
	pbeg = text; pend = top;
			
	{ //Reads the number of words of the vocabulary (stored in the text file as a long value)
		unsigned int numbertmp;
		int z;
		unsigned char *c;
		c= (unsigned char*) &numbertmp;
		for (z=1; z<=4;z++) {
			*c =*pbeg;
			c++;
			pbeg++;
		}
		fprintf(stderr,"\n\t*Number of words in the vocabulary is: %ld",(long)numbertmp);
	
		initialize_hash(numbertmp); //initializes decompressor for up to N_value words
	}			
	
	/*Reads one word at a time (j) and adds it to the vocabulary vector.*/
	{	unsigned long j;
		unsigned long RELPOS = 0;
		j=0;
		while(1){
		
			while ( pbeg < pend )   {
				//reads the word
				codeStart = pbeg;
				while (*pbeg !='\0') {
					pbeg++;
				}
				
				//adds the word to the vocabulary
				{	
					unsigned long addr;
	
					register unsigned long len;
					unsigned char *w;
					register int i;
			
					len = pbeg-codeStart;
					j = search ((unsigned char *)codeStart, len, &addr);			
											
					//adds the word to the hash Table

					insertElement( codeStart, len,&addr);

					hash[addr].len=len;
					{	unsigned long aCode;   //generates the codeword of the word
						getCode(RELPOS,&(aCode), &(hash[addr].size));						
						getBytes(hash[addr].codeword, aCode, hash[addr].size);	
						//hash[addr].codeword = ; hash[addr].len= ;
					}
					RELPOS ++;
				}
			
				j++;
				pbeg++;
			}
			
		nextbuf=pbeg;
		
		bufLoad (V, (char*)nextbuf);
		if (bufEmpty(V)) break;
		bufCurrent (V,(byte**)&text,(byte**)&top);
		pbeg = text; pend = top;
		}
	}
	
	bufDestroy(V);
	close(v);
	free(voc);		
}

/*-----------------------------------------------------------------------
 Loads the next "pattern" from the file of patterns, and stores it into
  structure nextwordpattern (global variable)
 ------------------------------------------------------------------------*/
void LoadPatternsFile(FILE *file, unsigned long numPats){

    unsigned long int i;
    unsigned long int numBytesWord;
    unsigned long int frequency;

	//FORMAT OF THE FILE ...
	// [WORD]
	// %s\n"

	for (i=0;i<numPats;i++) {
		fscanf(file,"%s\n", nextwordpattern.word);
		//nextwordpattern.word[numBytesWord]='\0';
		nextwordpattern.numBytesWord = strlen(nextwordpattern.word);
	}
}


void showMessage() {
	printf("\n* Horspool searcher for End-Tagged Dense Code v0.9 (ETDC)[semistatic compressor]");
	printf("\n* CopyRight(c)2006  Nieves R. Brisaboa, Gonzalo Navarro, Antonio Faria and Jos R. Param\n");
	
    printf("\n  #This program comes with ABSOLUTELY NO WARRANTY; ");
    printf("\n  #This is free software, and you are welcome to redistribute it ");
    printf("\n  #under certain conditions\n\n");
}

//*********************************************************************
//*********************************************************************
//*********************************************************************

int main (int argc, char *argv[]) {
	unsigned int z;
	unsigned char nombreFichTiempos[100];

	unsigned int iterations = 1;
	unsigned int iterActual =1;


	unsigned long pos,i;
	
	tTiempo miTiempo1;
	tTiempo miTiempo2;

	tTiempo miTiempo11;
	tTiempo miTiempo12;

	byte *pattern;   //patterns to be searched for (in each iteration)
	int len;   	   //len of the patterns searched in each iteration

	double totalTime;
	double totalVarianza;
  	totalTime =0.0;
  	totalVarianza = 0.0;

  	showMessage();
  	if (argc < 5) {
		printf("\nUsage: %s <compressedFile> <patternsFile> <iterations> <FTIEMPOS> \n\n", *argv);
		exit(0);
	}

	//prepares the file that will contain the measured times.
	nombreFichTiempos[0]='\0';
	strcat(nombreFichTiempos,argv[4]);
	strcat(nombreFichTiempos,"_.m");
	if ((ftiempos = fopen(nombreFichTiempos,"w")) == NULL)	{
		  fprintf(stderr,
			  "Error opening file for times: %s\n", nombreFichTiempos);
		  exit(0);
	}

	getTiempo(&miTiempo1);
	getTiempo(&miTiempo2);

	//process parameters
	iterations = atoi(argv[3]);

	fprintf(stderr, "fichero de tiempos %s\n", nombreFichTiempos);
	fprintf(stderr, "fichero comprimido: [%s],  fichero vocabulario: [%s]\n", argv[1], argv[2]);
	fprintf(stderr, "iterations [%d]\n", iterations);

	_memMgr = createMemoryManager();

	//opens the file containing the random patterns
	if ((fPatterns = fopen(argv[2],"r")) == NULL)
	{
	  fprintf(stderr,"\nError en la apertura del fichero %s\n", argv[2]);
	  exit(0);
	}
	
		
	//loads the vocabulary of the compressor into a hash table that will be used to obtain
	//the codewords associated to a search pattern (a word), previously to perform a search.
	loadCompressorHashTable(argv[1]);


	//FOR LOOP OF ITERATIONS STARTS. ******************
	getTiempo(&miTiempo11);
	getTiempo(&miTiempo1);

	// iterations ... *****************************************************
	
	for (;iterActual <= iterations; iterActual++) {
		
		fprintf(stderr,"\n Search: %d of %d",iterActual,iterations);
		
		//Read next random pattern...
		fprintf(stderr, "\n READING NEXT RANDOM PATTERN... \n");
		
		LoadPatternsFile(fPatterns,1);   //loads the next ("1") pattern from the "patterns file"
								
		//Once the compressed patterns is known, we search for it into the file.
		//  Time measures take only into account the time needed to search for each pattern inside the hash table
		//  of patterns, and the traversing the full file searching for the compressed search patterns. 
		//  It skips times needed to load the vocabulary, the hash table, ...
		/*********************/
		getTiempo(&miTiempo1);
		
		{//Searches for the loaded patterns inside the hash table 
		 //	and obtains their codeword and codeword lenght.
			unsigned long addr;
					
			z=search (nextwordpattern.word, nextwordpattern.numBytesWord, &addr);	
			if (z){
				fprintf(stderr, "\n  chosen word: [%s],  codesize [%ld]\n ",nextwordpattern.word,(long)hash[addr].size);
				pattern= (byte *) hash[addr].codeword;
				len = hash[addr].size;			
			}
			else
				printf("\n word \"%s\" not found in the vocabulary!\n",nextwordpattern.word);
		}

		//searches inside the compressed file.
		if (z)
			searchFullFile(argv[1],pattern,len);

		getTiempo(&miTiempo2);
		/*********************/
		
		totalTime  += miTiempo2.user - miTiempo1.user;	
		totalVarianza += ((miTiempo2.user - miTiempo1.user) * (miTiempo2.user - miTiempo1.user));
		
		fprintf(ftiempos,"%s(%d) = %f ;\n", argv[4],(iterActual) , miTiempo2.user - miTiempo1.user);
		fflush(ftiempos);
		muestraTiempo(miTiempo1,miTiempo2,"SEARCH TIME");
	
		
	} //end for iterations. *****************************************************

	getTiempo(&miTiempo12);

	
	//Gives some statistics of the current run...
	fprintf(stderr,"\n AVG USER TIME (only search process) = %f sec.",totalTime/iterations);
	{	double varr;		
		varr= (totalVarianza / iterations) - ((totalTime / iterations) *(totalTime / iterations));
		fprintf(stderr,"\n Desv est : %f  ", sqrt(varr * iterations / (iterations -1))  );
		fprintf(stderr,"\n Varianza : %f  ", varr  );
		fprintf(stderr,"\n Desv tp : %f  \n\n", sqrt(varr)  );
		
			
		fprintf(ftiempos,"mean(%s) %% %f \n", argv[4],totalTime/iterations );
		fprintf(ftiempos,"std(%s)  %% %f \n", argv[4],sqrt(varr * iterations / (iterations -1)) );
		fprintf(ftiempos,"%s=sort(%s);\n", argv[4],argv[4]  );
		fprintf(ftiempos,"plot(%s,'b')\n", argv[4]);
	}
	
	//Frees memory, closes file handlers ...........
	fclose (ftiempos);
	fclose (fPatterns);
	//destroyMemoryManager(_memMgr);

	freeHashTable();
}
